
#include "ast.h"

#include <cassert>
#include <iostream>
#include <map>
#include <stack>

#include <llvm/Constants.h>
#include <llvm/Instructions.h>
#include <llvm/Module.h>
#include <llvm/DerivedTypes.h>
#include <llvm/Analysis/Verifier.h>
#include <llvm/ExecutionEngine/GenericValue.h>
#include <llvm/ExecutionEngine/Interpreter.h>
#include <llvm/ExecutionEngine/JIT.h>
#include <llvm/ModuleProvider.h>
#include <llvm/Support/Streams.h>


extern FILE *yyin;

TranslationUnitAST *yyparse(const std::string &filename);
const std::string *intern(const std::string &);


class GenLLVM: protected ASTVisitor {
  llvm::Module *M;
  std::stack<llvm::Value*> temps;
  unsigned want_reference: 1;
  unsigned want_function: 1;

public:
  GenLLVM(llvm::Module *M):
    M(M) {}

  llvm::Module *operator()(AST *ast) {
    using namespace llvm;

    current_scope = global_scope = new Scope;

    assert(temps.empty());

    current_basic_block = 0;
    current_function = 0;
    want_reference = false;
    want_function = false;

    ast->accept(this);

    delete current_scope;

    return M;
  }

protected:
  using ASTVisitor::visit;
  using ASTVisitor::endVisit;

  llvm::Function *current_function;
  llvm::BasicBlock *current_basic_block;

  class TypeCompiler: protected ASTVisitor {
    const std::string *id;
    const llvm::Type *llvm_type;

  public:
    TypeCompiler() {}

    std::pair<const std::string*, const llvm::Type *> operator()(AST *ast, const llvm::Type *ty) {
      id = 0;
      llvm_type = ty;

      AST::acceptNode(ast, this);

      return std::make_pair(id, llvm_type);
    }

  protected:
    using ASTVisitor::visit;
    using ASTVisitor::endVisit;

    virtual bool visit(IdentifierDeclaratorAST *ast) {
      id = ast->id;
      return true;
    }

    virtual bool visit(PointerAST *) {
      llvm_type = llvm::PointerType::get(llvm_type);
      return true;
    }

    virtual bool visit(FunctionDeclaratorAST *ast) {
      using namespace llvm;

      std::vector<const llvm::Type*> args;

      TypeCompiler type_cc;

      if (ArgumentsAST *arguments = ast->arguments)
	{
	  args.resize(arguments->size());

	  for (size_t index = 0; index < arguments->size(); ++index)
	    {
	      ArgumentAST *arg = arguments->at(index);

	      std::pair<const std::string *, const Type *> r = type_cc(arg, typeSpecifier(arg->type_specifier));
	      args[index] = r.second;
	    }
	}

      llvm_type = llvm::FunctionType::get(llvm_type, args, false); 
      return true;
    }

    virtual bool visit(ArrayDeclaratorAST *ast) {
      llvm_type = llvm::ArrayType::get(llvm_type, ast->dimension);
      return true;
    }
  };

  TypeCompiler type_cc;

  static const llvm::Type *typeSpecifier(const std::string *s) {
    using namespace llvm;

    const Type *ty = Type::Int32Ty;

    if (*s == "char")
      ty = Type::Int8Ty;

    return ty;
  }

  virtual bool visit(DeclarationAST *ast) {
    processDeclarators(ast->declarators, typeSpecifier(ast->type_specifier));
    return false;
  }

  virtual void processDeclarators(DeclaratorsAST *ast, const llvm::Type *ty) {
    using namespace llvm;

    for (DeclaratorsAST::iterator it = ast->begin(); it != ast->end(); ++it) {
      std::pair<const std::string*, const Type*> r = type_cc(*it, ty);
      const std::string *id = r.first;
      const Type *ty = r.second;

      assert(id != 0);

      assert(current_scope != 0);

      if (! current_function)
	{
	  Value *value = 0;

	  if (isa<FunctionType>(ty))
	    value = M->getOrInsertFunction(*id, cast<FunctionType>(ty));
	  else
	    {
	      value = new GlobalVariable(ty, false, 
					 GlobalVariable::ExternalLinkage,
					 Constant::getNullValue(ty), *id);
	    }

	  current_scope->insert(id, value);
	}
      else
	{
	  assert(current_basic_block != 0);

	  Value *value = new AllocaInst(ty, *id, current_basic_block);
	  current_scope->insert(id, value);
	}
    }
  }

  virtual void endVisit(ExpressionStatementAST *ast) {
    if (ast->expression)
      temps.pop();
  }

  virtual bool visit(CastExpressionAST *ast) {
    using namespace llvm;

    const Type *ty = type_cc(ast->declarator, typeSpecifier(ast->type_specifier)).second;

    AST::acceptNode(ast->expression, this);
    Value *v = temps.top();
    temps.pop();

    Instruction::CastOps castOp = CastInst::getCastOpcode(v, true, ty, true);
    temps.push(CastInst::create(castOp, v, ty, "", current_basic_block));

    return false;
  }

  virtual bool visit(WhileStatementAST *ast) {
    using namespace llvm;

    BasicBlock *top_basic_block = new BasicBlock("top", current_function);
    BasicBlock *body_basic_block = new BasicBlock("body", current_function);
    BasicBlock *end_basic_block = new BasicBlock("end", current_function);

    new BranchInst(top_basic_block, current_basic_block);

    current_basic_block = top_basic_block;
    AST::acceptNode(ast->condition, this);
    
    Value *cond = temps.top();
    temps.pop();

    if (cond->getType() != Type::Int1Ty)
      {
	assert(isa<IntegerType>(cond->getType()));

	Value *Zero = ConstantInt::get(cond->getType(), 0);
	cond = new ICmpInst(ICmpInst::ICMP_NE, cond, Zero, "", current_basic_block);
      }

    new BranchInst(body_basic_block, end_basic_block, cond, current_basic_block);

    current_basic_block = body_basic_block;
    AST::acceptNode(ast->statement, this);

    if (! current_basic_block->getTerminator())
      new BranchInst(top_basic_block, current_basic_block);

    current_basic_block = end_basic_block;

    return false;
  }

  virtual bool visit(IfStatementAST *ast) {
    using namespace llvm;

    BasicBlock *then_basic_block = new BasicBlock("then", current_function);
    BasicBlock *else_basic_block = new BasicBlock("else", current_function);
    BasicBlock *end_basic_block = new BasicBlock("end", current_function);

    AST::acceptNode(ast->condition, this);

    Value *Zero = ConstantInt::get(Type::Int32Ty, 0);

    Value *cond = temps.top();
    temps.pop();

    if (cond->getType() != Type::Int1Ty)
      {
	assert(isa<IntegerType>(cond->getType()));

	Value *Zero = ConstantInt::get(cond->getType(), 0);
	cond = new ICmpInst(ICmpInst::ICMP_NE, cond, Zero, "", current_basic_block);
      }

    new BranchInst(then_basic_block, else_basic_block, cond, current_basic_block);

    current_basic_block = then_basic_block;
    AST::acceptNode(ast->then_statement, this);
    if (! current_basic_block->getTerminator())
      new BranchInst(end_basic_block, current_basic_block);

    current_basic_block = else_basic_block;
    AST::acceptNode(ast->else_statement, this);
    if (! current_basic_block->getTerminator())
      new BranchInst(end_basic_block, current_basic_block);

    current_basic_block = end_basic_block;

    return false;
  }

  class ProcessFunctionDeclarator: protected ASTVisitor {
    std::vector<const llvm::Type *> *args;
    std::vector<const std::string *> *arg_names;

  public:
    bool operator()(DeclaratorAST *ast, std::vector<const llvm::Type *> *args, std::vector<const std::string *> *arg_names) {
      this->args = args;
      this->arg_names = arg_names;
      ast->accept(this);
      return true;
    }

  protected:
    using ASTVisitor::visit;
    using ASTVisitor::endVisit;

    virtual bool visit(ArgumentsAST *arguments) {
      using namespace llvm;

      TypeCompiler type_cc;
      size_t argc = arguments->size();

      for (size_t index = 0; index < argc; ++index)
	{
	  ArgumentAST *arg = arguments->at(index);

	  std::pair<const std::string *, const Type *> r = type_cc(arg->declarator, typeSpecifier(arg->type_specifier));

	  assert(r.first != 0);

	  arg_names->push_back(r.first);
	  args->push_back(r.second);
	}

      return false;
    }
  };

  ProcessFunctionDeclarator processFunctionDeclarator;

  virtual bool visit(FunctionAST *ast) {
    using namespace llvm;

    std::vector<const Type*> args;
    std::vector<const std::string*> arg_names;

    assert(current_scope != 0);

    current_scope = new Scope(current_scope);

    std::pair<const std::string *, const Type *> r = type_cc(ast->declarator, typeSpecifier(ast->type_specifier));

    const FunctionType *ty = cast<FunctionType>(r.second);
    assert(ty != 0);

    processFunctionDeclarator(ast->declarator, &args, &arg_names);

    current_function = new Function(ty, Function::ExternalLinkage, *r.first, M);

    current_scope->previous->insert(r.first, current_function);

    current_basic_block = new BasicBlock("", current_function);

    Function::arg_iterator arg_it = current_function->arg_begin();
    for (size_t index = 0; index < arg_names.size(); ++index, ++arg_it)
      {
	assert(arg_it != current_function->arg_end());

	const Type *ty = args.at(index);

	const std::string *id = arg_names.at(index);
	assert(id != 0);

	Value *where = new AllocaInst(ty, *id, current_basic_block);
	new StoreInst(arg_it, where, false, current_basic_block);
	current_scope->insert(id, where);
      }

    AST::acceptNode(ast->block, this);

    return false;
  }

  virtual void endVisit(FunctionAST *ast) {
    using namespace llvm;

    if (! current_basic_block->getTerminator())
      new ReturnInst(ConstantInt::get(Type::Int32Ty, 0), current_basic_block);

    Scope *scope = current_scope;
    current_scope = current_scope->previous;
    delete scope;
    current_function = 0;
  }

  virtual bool visit(IntConstantAST *ast) {
    using namespace llvm;

    Value *value = ConstantInt::get(Type::Int32Ty, ast->value);
    temps.push(value);
  }

  llvm::Value *findOrCreateStringLiteral(const std::string *s) {
    using namespace llvm;

    std::map<const std::string *, Value *>::iterator it = string_literal_table.find(s);

    if (it == string_literal_table.end())
      {
	Constant *initializer = ConstantArray::get(*s);

	Value *g = new GlobalVariable(initializer->getType(), /*constant=*/ true,
				      GlobalVariable::ExternalLinkage,
				      initializer, "", M);

	string_literal_table.insert(std::make_pair(s, g));

	return g;
      }

    return it->second;
  }

  virtual bool visit(StringLiteralAST *ast) {
    using namespace llvm;

    assert(current_basic_block != 0);

    ConstantInt *Zero = ConstantInt::get(Type::Int32Ty, 0);

    Value *value = new GetElementPtrInst(findOrCreateStringLiteral(ast->value),
					 Zero, Zero,
					 "", current_basic_block);

    assert(isa<PointerType>(value->getType()));

    temps.push(value);
  }

  virtual bool visit(IdentifierAST *ast) {
    using namespace llvm;

    Value *value = resolve(ast->id);

    if (! value && want_function)
      {
	std::vector<const Type*> noargs;
	FunctionType *ty = FunctionType::get(Type::Int32Ty, noargs, true);
	value = M->getOrInsertFunction(*ast->id, ty);
	global_scope->insert(ast->id, value);
      }

    assert(value != 0);

    if (! isa<Function>(value))
      {
	assert(isa<PointerType>(value->getType()));

	if (! want_reference)
	  value = new LoadInst(value, "", current_basic_block);
      }

    temps.push(value);

    return true;
  }

  virtual bool visit(BlockAST *ast) {
    current_scope = new Scope(current_scope);
    return true;
  }

  virtual void endVisit(BlockAST *ast) {
    Scope *scope = current_scope;
    current_scope = current_scope->previous;
    delete scope;
  }

  bool wantReference(bool b) {
    bool was = want_reference;
    want_reference = b;
    return was;
  }

  bool wantFunction(bool b) {
    bool was = want_function;
    want_function = b;
    return was;
  }

  virtual bool visit(CallExpressionAST *ast) {
    using namespace llvm;

    bool was = wantFunction(true);
    AST::acceptNode(ast->expression, this);
    wantFunction(was);

    Value *f = temps.top(); temps.pop();

    assert(isa<Function>(f));

    if (ast->expression_list && ast->expression_list->size())
      {
	size_t argc = ast->expression_list->size();

	Value **args = new Value* [argc];
	for (size_t i = 0; i < ast->expression_list->size(); ++i)
	  {
	    ExpressionAST *expr = ast->expression_list->at(i);
	    AST::acceptNode(expr, this);

	    args[i] = temps.top();
	    temps.pop();
	  }

	temps.push(new CallInst(f, args, int(argc), "", current_basic_block));
	delete[] args;
      }
    else
      temps.push(new CallInst(f, "", current_basic_block));

    return false;
  }

  virtual bool visit(UnaryExpressionAST *ast) {
    using namespace llvm;

    bool lvalue = wantReference(ast->op == op_ref);
    AST::acceptNode(ast->expression, this);
    wantReference(lvalue);

    Value *v = temps.top();
    temps.pop();

    switch (ast->op) {

    case op_ref: {
      assert(isa<PointerType>(v->getType()));
      temps.push(v);
    } break;

    case op_deref: {
      assert(isa<PointerType>(v->getType()));
      if (! want_reference)
	v = new LoadInst(v, "", current_basic_block);
      temps.push(v);
    } break;

    case op_neg: {
      if (isa<PointerType>(v->getType()))
	v = CastInst::createPointerCast(v, Type::Int32Ty, "", current_basic_block);

      assert(isa<IntegerType>(v->getType()));

      temps.push(BinaryOperator::createNeg(v, "", current_basic_block));
    } break;

    case op_not: {
      if (isa<PointerType>(v->getType()))
	v = CastInst::createPointerCast(v, Type::Int32Ty, "", current_basic_block);

      assert(isa<IntegerType>(v->getType()));

      temps.push(new ICmpInst(ICmpInst::ICMP_EQ, v, ConstantInt::get(v->getType(), 0),
			      "", current_basic_block));
    } break;

    } // switch

    return false;
  }

  virtual bool visit(BinaryExpressionAST *ast) {
    using namespace llvm;

    bool lvalue = wantReference(ast->op == op_assign);
    AST::acceptNode(ast->left_expression, this);
    wantReference(false);

    assert(! temps.empty());
    Value *v1 = temps.top(); temps.pop();

    AST::acceptNode(ast->right_expression, this);
    wantReference(lvalue);

    assert(! temps.empty());
    Value *v2 = temps.top(); temps.pop();

    const Type *t1 = v1->getType();
    const Type *t2 = v2->getType();

    switch (ast->op) {
    case op_add: {
      if (isa<IntegerType>(t1) && isa<IntegerType>(t2))
	{
	  if (t1->getPrimitiveSizeInBits() > t2->getPrimitiveSizeInBits())
	    v2 = CastInst::createIntegerCast(v2, t1, true, "", current_basic_block);

	  else if (t1->getPrimitiveSizeInBits() != t2->getPrimitiveSizeInBits())
	    v1 = CastInst::createIntegerCast(v1, t2, true, "", current_basic_block);

	  temps.push(BinaryOperator::createAdd(v1, v2, "", current_basic_block));
	}
      else if (isa<PointerType>(t1))
	{
	  assert(isa<IntegerType>(t2));

	  temps.push(new GetElementPtrInst(v1, v2, "", current_basic_block));
	}
      else if (isa<PointerType>(t2))
	{
	  assert(isa<IntegerType>(t1));

	  temps.push(new GetElementPtrInst(v2, v1, "", current_basic_block));
	}
      else
	{
	  std::cerr << "*** Internal Compiler Errror." << std::endl;
	  exit(EXIT_FAILURE);
	}
    } break;

    case op_sub: {
      if (isa<IntegerType>(t1) && isa<IntegerType>(t2))
	{
	  if (t1->getPrimitiveSizeInBits() > t2->getPrimitiveSizeInBits())
	    v2 = CastInst::createIntegerCast(v2, t1, true, "", current_basic_block);

	  else if (t1->getPrimitiveSizeInBits() != t2->getPrimitiveSizeInBits())
	    v1 = CastInst::createIntegerCast(v1, t2, true, "", current_basic_block);

	  temps.push(BinaryOperator::createSub(v1, v2, "", current_basic_block));
	}
      else if (isa<PointerType>(t1))
	{
	  if (t1 == t2)
	    {
	      v1 = CastInst::createPointerCast(v1, Type::Int32Ty, "", current_basic_block);
	      v2 = CastInst::createPointerCast(v2, Type::Int32Ty, "", current_basic_block);

	      temps.push(new IntToPtrInst(BinaryOperator::createSub(v1, v2, "", current_basic_block), t1, "", current_basic_block));
	    }
	  else
	    {
	      assert(isa<IntegerType>(t2));

	      temps.push(new GetElementPtrInst(v1, BinaryOperator::createNeg(v2, "", current_basic_block),
					       "", current_basic_block));
	    }
	}
      else
	{
	  std::cerr << "*** Internal Compiler Errror." << std::endl;
	  exit(EXIT_FAILURE);
	}
    } break;

    case op_mul:
      assert(isa<IntegerType>(t1) && isa<IntegerType>(t2));

      if (t1->getPrimitiveSizeInBits() > t2->getPrimitiveSizeInBits())
	v2 = CastInst::createIntegerCast(v2, t1, true, "", current_basic_block);

      else if (t1->getPrimitiveSizeInBits() != t2->getPrimitiveSizeInBits())
	v1 = CastInst::createIntegerCast(v1, t2, true, "", current_basic_block);

      temps.push(BinaryOperator::createMul(v1, v2, "", current_basic_block));
      break;

    case op_div:
      assert(isa<IntegerType>(t1) && isa<IntegerType>(t2));

      if (t1->getPrimitiveSizeInBits() > t2->getPrimitiveSizeInBits())
	v2 = CastInst::createIntegerCast(v2, t1, true, "", current_basic_block);

      else if (t1->getPrimitiveSizeInBits() != t2->getPrimitiveSizeInBits())
	v1 = CastInst::createIntegerCast(v1, t2, true, "", current_basic_block);

      temps.push(BinaryOperator::createSDiv(v1, v2, "", current_basic_block));
      break;

    case op_assign: {
      const PointerType *ptr = cast<PointerType>(t1);
      assert(ptr != 0);

      const Type *t = ptr->getElementType();

      if (t != t2)
	{
	  Instruction::CastOps castOp = CastInst::getCastOpcode(v2, true, t, true);
	  v2 = CastInst::create(castOp, v2, t, "", current_basic_block);
	}

      new StoreInst(v2, v1, false, current_basic_block);
      temps.push(v2);
    } break;

    case op_eq: case op_not_eq:
    case op_lt: case op_gt:
    case op_le: case op_ge: {
      if (isa<PointerType>(t1))
	v1 = CastInst::createPointerCast(v1, Type::Int32Ty, "", current_basic_block);

      if (isa<PointerType>(t2))
	v2 = CastInst::createPointerCast(v2, Type::Int32Ty, "", current_basic_block);

      assert(isa<IntegerType>(v1->getType()));
      assert(isa<IntegerType>(v2->getType()));

      if (v1->getType() != v2->getType())
	{
	  v1 = CastInst::createIntegerCast(v1, Type::Int32Ty, true, "", current_basic_block);

	  v2 = CastInst::createIntegerCast(v2, Type::Int32Ty, true, "", current_basic_block);
	}

      temps.push(new ICmpInst(getICmpOpcode(ast->op), v1, v2, "", current_basic_block));
    } break;

    case op_or:
    case op_and:
    default:
      std::cerr << "*** Internal Compiler Error. Operator not implemented" << std::endl;
      exit(1);
      break;
    }

    return false;
  }

  static llvm::ICmpInst::Predicate getICmpOpcode(int op) {
    using namespace llvm;

    switch (op) {

    case op_eq:
      return ICmpInst::ICMP_EQ;

    case op_not_eq:
      return ICmpInst::ICMP_NE;

    case op_lt:
      return ICmpInst::ICMP_SLT;

    case op_gt:
      return ICmpInst::ICMP_SGT;

    case op_le:
      return ICmpInst::ICMP_SLE;

    case op_ge:
      return ICmpInst::ICMP_SGE;

    default:
      std::cerr << "*** Internal Compiler Error." << std::endl;
      exit(EXIT_FAILURE);

    } // switch
  }

  virtual void endVisit(ReturnStatementAST *ast) {
    using namespace llvm;

    if (! ast->expression)
      new ReturnInst(current_basic_block);
    else
      {
	Value *r = temps.top();
	temps.pop();
	new ReturnInst(r, current_basic_block);
      }
  }

  llvm::Value *resolve(const std::string *id) {
    return current_scope->resolve(id);
  }

  std::map<const std::string *, llvm::Value*> string_literal_table;

  struct Scope {
    Scope *previous;
    std::map<const std::string *, llvm::Value*> m;

    Scope(Scope *previous = 0):
      previous(previous) {}

    void insert(const std::string *id, llvm::Value *v) {
      assert(m.find(id) == m.end());
      m[id] = v;
    }

    llvm::Value *resolve(const std::string *id) {
      std::map<const std::string *, llvm::Value*>::iterator it = m.find(id);

      if (it != m.end())
	return it->second;

      else if (previous)
	return previous->resolve(id);

      return 0;
    }
  };

  Scope *current_scope;
  Scope *global_scope;
};

int main(int, char *argv[])
{
  using namespace llvm;

  if (! argv[1])
    {
      std::cerr << argv[0] << ": no input file" << std::endl;
      exit(EXIT_FAILURE);
    }

  std::string yyfilename = argv[1];

  yyin = fopen(yyfilename.c_str(), "r");
  if (! yyin)
    {
      std::cerr << argv[0] << ": file not found" << std::endl;
      exit(EXIT_FAILURE);
    }

  TranslationUnitAST *translation_unit = yyparse(yyfilename);
  if (! translation_unit)
    exit(EXIT_FAILURE);

  Module *M = new Module(yyfilename);
  GenLLVM genLLVM(M);
  genLLVM(translation_unit);
  
#if 0
  M->print(&std::cout);
#endif

#if 0
  verifyModule(*M);
#endif

#if 0
  WriteBytecodeToFile(M, cout);
#endif

#if 1
  Function *main_function = M->getFunction("main");
  assert(main_function != 0);

  // interpreter
  ExistingModuleProvider *MP = new ExistingModuleProvider(M);
  ExecutionEngine *EE = ExecutionEngine::create(MP, false);

  std::vector<GenericValue> args;

  GenericValue result = EE->runFunction(main_function, args);
  // std::cout << result.IntVal.toString() << std::endl;
#endif

  fclose(yyin);

  exit(EXIT_SUCCESS);
}
